@@ -12,13 +12,13 @@ use core::any::{type_name, TypeId};
 use core::hash::{BuildHasher, BuildHasherDefault, Hasher};
 use core::ops::Deref;
 use core::ptr::{self, NonNull};
-use core::{fmt, mem, slice};
+use core::{fmt, slice};
 
 use hashbrown::{hash_map::DefaultHashBuilder, HashMap};
 
 use crate::borrow::AtomicBorrow;
 use crate::query::Fetch;
-use crate::{align, Access, Component, Query};
+use crate::{Access, Component, Query};
 
 /// A collection of entities having the same component types
 ///
@@ -26,11 +26,11 @@ use crate::{align, Access, Component, Query};
 /// [`World`](crate::World).
 pub struct Archetype {
     types: Vec<TypeInfo>,
-    state: OrderedTypeIdMap<TypeState>,
+    index: OrderedTypeIdMap<usize>,
     len: u32,
     entities: Box<[u32]>,
-    data: NonNull<u8>,
-    data_size: usize,
+    /// One allocation per type, in the same order as `types`
+    data: Box<[Data]>,
     /// Maps static bundle types to the archetype that an entity from this archetype is moved to
     /// after removing the components from that bundle.
     pub(crate) remove_edges: TypeIdMap<u32>,
@@ -58,25 +58,27 @@ impl Archetype {
     pub(crate) fn new(types: Vec<TypeInfo>) -> Self {
         let max_align = types.first().map_or(1, |ty| ty.layout.align());
         Self::assert_type_info(&types);
+        let component_count = types.len();
         Self {
-            state: OrderedTypeIdMap::new(types.iter().map(|ty| (ty.id, TypeState::new(0)))),
+            index: OrderedTypeIdMap::new(types.iter().enumerate().map(|(i, ty)| (ty.id, i))),
             types,
             entities: Box::new([]),
             len: 0,
-            data: NonNull::new(max_align as *mut u8).unwrap(),
-            data_size: 0,
+            data: (0..component_count)
+                .map(|_| Data {
+                    state: AtomicBorrow::new(),
+                    storage: NonNull::new(max_align as *mut u8).unwrap(),
+                })
+                .collect(),
             remove_edges: HashMap::default(),
         }
     }
 
     pub(crate) fn clear(&mut self) {
-        for ty in &self.types {
+        for (ty, data) in self.types.iter().zip(&*self.data) {
             for index in 0..self.len {
                 unsafe {
-                    let removed = self
-                        .get_dynamic(ty.id, ty.layout.size(), index)
-                        .unwrap()
-                        .as_ptr();
+                    let removed = data.storage.as_ptr().add(index as usize * ty.layout.size());
                     (ty.drop)(removed);
                 }
             }
@@ -91,21 +93,22 @@ impl Archetype {
 
     /// Whether this archetype contains components with the type identified by `id`
     pub fn has_dynamic(&self, id: TypeId) -> bool {
-        self.state.contains_key(&id)
+        self.index.contains_key(&id)
     }
 
     /// Find the state index associated with `T`, if present
     pub(crate) fn get_state<T: Component>(&self) -> Option<usize> {
-        self.state.search(&TypeId::of::<T>())
+        self.index.get(&TypeId::of::<T>()).copied()
     }
 
     /// Get the address of the first `T` component using an index from `get_state::<T>`
     pub(crate) fn get_base<T: Component>(&self, state: usize) -> NonNull<T> {
-        let (id, state) = self.state.get_from_index(state);
-        assert_eq!(id, &TypeId::of::<T>());
+        assert_eq!(self.types[state].id, TypeId::of::<T>());
 
         unsafe {
-            NonNull::new_unchecked(self.data.as_ptr().add(state.offset).cast::<T>() as *mut T)
+            NonNull::new_unchecked(
+                self.data.get_unchecked(state).storage.as_ptr().cast::<T>() as *mut T
+            )
         }
     }
 
@@ -124,35 +127,29 @@ impl Archetype {
     }
 
     pub(crate) fn borrow<T: Component>(&self, state: usize) {
-        let (id, state) = self.state.get_from_index(state);
-        assert_eq!(id, &TypeId::of::<T>());
+        assert_eq!(self.types[state].id, TypeId::of::<T>());
 
-        if !state.borrow.borrow() {
+        if !self.data[state].state.borrow() {
             panic!("{} already borrowed uniquely", type_name::<T>());
         }
     }
 
     pub(crate) fn borrow_mut<T: Component>(&self, state: usize) {
-        let (id, state) = self.state.get_from_index(state);
-        assert_eq!(id, &TypeId::of::<T>());
+        assert_eq!(self.types[state].id, TypeId::of::<T>());
 
-        if !state.borrow.borrow_mut() {
+        if !self.data[state].state.borrow_mut() {
             panic!("{} already borrowed", type_name::<T>());
         }
     }
 
     pub(crate) fn release<T: Component>(&self, state: usize) {
-        let (id, state) = self.state.get_from_index(state);
-        assert_eq!(id, &TypeId::of::<T>());
-
-        state.borrow.release();
+        assert_eq!(self.types[state].id, TypeId::of::<T>());
+        self.data[state].state.release();
     }
 
     pub(crate) fn release_mut<T: Component>(&self, state: usize) {
-        let (id, state) = self.state.get_from_index(state);
-        assert_eq!(id, &TypeId::of::<T>());
-
-        state.borrow.release_mut();
+        assert_eq!(self.types[state].id, TypeId::of::<T>());
+        self.data[state].state.release_mut();
     }
 
     /// Number of entities in this archetype
@@ -213,8 +210,10 @@ impl Archetype {
         debug_assert!(index <= self.len);
         Some(NonNull::new_unchecked(
             self.data
+                .get_unchecked(*self.index.get(&ty)?)
+                .storage
                 .as_ptr()
-                .add(self.state.get(&ty)?.offset + size * index as usize)
+                .add(size * index as usize)
                 .cast::<u8>(),
         ))
     }
@@ -256,71 +255,66 @@ impl Archetype {
     fn grow_exact(&mut self, increment: u32) {
         unsafe {
             let old_count = self.len as usize;
+            let old_cap = self.entities.len();
             let new_cap = self.entities.len() + increment as usize;
             let mut new_entities = vec![!0; new_cap].into_boxed_slice();
             new_entities[0..old_count].copy_from_slice(&self.entities[0..old_count]);
             self.entities = new_entities;
 
-            let old_data_size = mem::replace(&mut self.data_size, 0);
-            let mut new_state =
-                OrderedTypeIdMap::new(self.types.iter().map(|ty| (ty.id, TypeState::new(0))));
-            for ty in &self.types {
-                self.data_size = align(self.data_size, ty.layout.align());
-                new_state.get_mut(&ty.id).unwrap().offset = self.data_size;
-                self.data_size += ty.layout.size() * new_cap;
-            }
-            let max_align = self.types.first().map_or(1, |x| x.layout.align());
-            let new_data = if self.data_size == 0 {
-                NonNull::new(max_align as *mut u8).unwrap()
-            } else {
-                NonNull::new(alloc(
-                    Layout::from_size_align(self.data_size, max_align).unwrap(),
-                ))
-                .unwrap()
-            };
-            if old_data_size != 0 {
-                for ty in &self.types {
-                    let old_off = self.state.get(&ty.id).unwrap().offset;
-                    let new_off = new_state.get(&ty.id).unwrap().offset;
-                    ptr::copy_nonoverlapping(
-                        self.data.as_ptr().add(old_off),
-                        new_data.as_ptr().add(new_off),
-                        ty.layout.size() * old_count,
-                    );
-                }
-                dealloc(
-                    self.data.as_ptr().cast(),
-                    Layout::from_size_align_unchecked(
-                        old_data_size,
-                        self.types.first().map_or(1, |x| x.layout.align()),
-                    ),
-                );
-            }
+            let new_data = self
+                .types
+                .iter()
+                .zip(&*self.data)
+                .map(|(info, old)| {
+                    let storage = if info.layout.size() == 0 {
+                        NonNull::new(info.layout.align() as *mut u8).unwrap()
+                    } else {
+                        let mem = alloc(
+                            Layout::from_size_align(
+                                info.layout.size() * new_cap,
+                                info.layout.align(),
+                            )
+                            .unwrap(),
+                        );
+                        ptr::copy_nonoverlapping(
+                            old.storage.as_ptr(),
+                            mem,
+                            info.layout.size() * old_count,
+                        );
+                        if old_cap > 0 {
+                            dealloc(
+                                old.storage.as_ptr(),
+                                Layout::from_size_align(
+                                    info.layout.size() * old_cap,
+                                    info.layout.align(),
+                                )
+                                .unwrap(),
+                            );
+                        }
+                        NonNull::new(mem).unwrap()
+                    };
+                    Data {
+                        state: AtomicBorrow::new(), // &mut self guarantees no outstanding borrows
+                        storage,
+                    }
+                })
+                .collect::<Box<[_]>>();
 
             self.data = new_data;
-            self.state = new_state;
         }
     }
 
     /// Returns the ID of the entity moved into `index`, if any
     pub(crate) unsafe fn remove(&mut self, index: u32, drop: bool) -> Option<u32> {
         let last = self.len - 1;
-        for ty in &self.types {
-            let removed = self
-                .get_dynamic(ty.id, ty.layout.size(), index)
-                .unwrap()
-                .as_ptr();
+        for (ty, data) in self.types.iter().zip(&*self.data) {
+            let removed = data.storage.as_ptr().add(index as usize * ty.layout.size());
             if drop {
                 (ty.drop)(removed);
             }
             if index != last {
-                ptr::copy_nonoverlapping(
-                    self.get_dynamic(ty.id, ty.layout.size(), last)
-                        .unwrap()
-                        .as_ptr(),
-                    removed,
-                    ty.layout.size(),
-                );
+                let moved = data.storage.as_ptr().add(last as usize * ty.layout.size());
+                ptr::copy_nonoverlapping(moved, removed, ty.layout.size());
             }
         }
         self.len = last;
@@ -339,20 +333,12 @@ impl Archetype {
         mut f: impl FnMut(*mut u8, TypeId, usize),
     ) -> Option<u32> {
         let last = self.len - 1;
-        for ty in &self.types {
-            let moved = self
-                .get_dynamic(ty.id, ty.layout.size(), index)
-                .unwrap()
-                .as_ptr();
-            f(moved, ty.id(), ty.layout().size());
+        for (ty, data) in self.types.iter().zip(&*self.data) {
+            let moved_out = data.storage.as_ptr().add(index as usize * ty.layout.size());
+            f(moved_out, ty.id(), ty.layout().size());
             if index != last {
-                ptr::copy_nonoverlapping(
-                    self.get_dynamic(ty.id, ty.layout.size(), last)
-                        .unwrap()
-                        .as_ptr(),
-                    moved,
-                    ty.layout.size(),
-                );
+                let moved = data.storage.as_ptr().add(last as usize * ty.layout.size());
+                ptr::copy_nonoverlapping(moved, moved_out, ty.layout.size());
             }
         }
         self.len -= 1;
@@ -391,15 +377,14 @@ impl Archetype {
     /// Component types must match exactly.
     pub(crate) unsafe fn merge(&mut self, mut other: Archetype) {
         self.reserve(other.len);
-        for info in &self.types {
-            let src_off = other.state.get(&info.id()).unwrap().offset;
-            let src = other.data.as_ptr().add(src_off);
-            let dst_off = self.state.get(&info.id()).unwrap().offset;
-            let dst = self
-                .data
+        for ((info, dst), src) in self.types.iter().zip(&*self.data).zip(&*other.data) {
+            dst.storage
                 .as_ptr()
-                .add(dst_off + self.len as usize * info.layout.size());
-            dst.copy_from_nonoverlapping(src, other.len as usize * info.layout.size())
+                .add(self.len as usize * info.layout.size())
+                .copy_from_nonoverlapping(
+                    src.storage.as_ptr(),
+                    other.len as usize * info.layout.size(),
+                )
         }
         self.len += other.len;
         other.len = 0;
@@ -419,20 +404,30 @@ impl Archetype {
 impl Drop for Archetype {
     fn drop(&mut self) {
         self.clear();
-        if self.data_size != 0 {
-            unsafe {
-                dealloc(
-                    self.data.as_ptr().cast(),
-                    Layout::from_size_align_unchecked(
-                        self.data_size,
-                        self.types.first().map_or(1, |x| x.layout.align()),
-                    ),
-                );
+        if self.entities.len() == 0 {
+            return;
+        }
+        for (info, data) in self.types.iter().zip(&*self.data) {
+            if info.layout.size() != 0 {
+                unsafe {
+                    dealloc(
+                        data.storage.as_ptr(),
+                        Layout::from_size_align_unchecked(
+                            info.layout.size() * self.entities.len(),
+                            info.layout.align(),
+                        ),
+                    );
+                }
             }
         }
     }
 }
 
+struct Data {
+    state: AtomicBorrow,
+    storage: NonNull<u8>,
+}
+
 /// A hasher optimized for hashing a single TypeId.
 ///
 /// TypeId is already thoroughly hashed, so there's no reason to hash it again.
@@ -497,28 +492,6 @@ impl<V> OrderedTypeIdMap<V> {
     fn get(&self, id: &TypeId) -> Option<&V> {
         self.search(id).map(move |idx| &self.0[idx].1)
     }
-
-    fn get_mut(&mut self, id: &TypeId) -> Option<&mut V> {
-        self.search(id).map(move |idx| &mut self.0[idx].1)
-    }
-
-    fn get_from_index(&self, idx: usize) -> &(TypeId, V) {
-        &self.0[idx]
-    }
-}
-
-struct TypeState {
-    offset: usize,
-    borrow: AtomicBorrow,
-}
-
-impl TypeState {
-    fn new(offset: usize) -> Self {
-        Self {
-            offset,
-            borrow: AtomicBorrow::new(),
-        }
-    }
 }
 
 /// Metadata required to store a component
