@@ -9,26 +9,23 @@ use crate::{Component, Entity, MissingComponent, Query, QueryOne};
 #[derive(Copy, Clone)]
 pub struct EntityRef<'a> {
     archetype: &'a Archetype,
-    generation: u32,
+    entity: Entity,
     index: u32,
 }
 
 impl<'a> EntityRef<'a> {
-    pub(crate) unsafe fn new(archetype: &'a Archetype, generation: u32, index: u32) -> Self {
+    pub(crate) unsafe fn new(archetype: &'a Archetype, entity: Entity, index: u32) -> Self {
         Self {
             archetype,
-            generation,
+            entity,
             index,
         }
     }
 
     /// Get the [`Entity`] handle associated with this entity
+    #[inline]
     pub fn entity(&self) -> Entity {
-        let id = self.archetype.entity_id(self.index);
-        Entity {
-            generation: self.generation,
-            id,
-        }
+        self.entity
     }
 
     /// Borrow the component of type `T`, if it exists
